/* Copyright (C) 2017 Mono Wireless Inc. All Rights Reserved.    *
 * Released under MW-SLA-*J,*E (MONO WIRELESS SOFTWARE LICENSE   *
 * AGREEMENT).                                                   */

#include <sercmd.h>
#include <arduino.h>

#define SERCMD_SYNC_1 0xA5
#define SERCMD_SYNC_2 0x5A

typedef enum {
	E_SERCMD_BINARY_EMPTY = 0,      //!< E_SERCMD_BINARY_EMPTY
	E_SERCMD_BINARY_READSYNC,       //!< E_SERCMD_BINARY_READSYNC
	E_SERCMD_BINARY_READLEN,        //!< E_SERCMD_BINARY_READLEN
	E_SERCMD_BINARY_READLEN2,       //!< E_SERCMD_BINARY_READLEN2
	E_SERCMD_BINARY_READPAYLOAD,    //!< E_SERCMD_BINARY_READPAYLOAD
	E_SERCMD_BINARY_READCRC,        //!< E_SERCMD_BINARY_READCRC
	E_SERCMD_BINARY_PLUS1,          //!< E_SERCMD_BINARY_PLUS1
	E_SERCMD_BINARY_PLUS2,          //!< E_SERCMD_BINARY_PLUS2
	E_SERCMD_BINARY_COMPLETE = 0x80,//!< E_SERCMD_BINARY_COMPLETE
	E_SERCMD_BINARY_ERROR = 0x81,   //!< E_SERCMD_BINARY_ERROR
	E_SERCMD_BINARY_CRCERROR = 0x82 //!< E_SERCMD_BINARY_CRCERROR
} teSerCmd_Binary;

typedef enum {
	E_SERCMD_ASCII_CMD_EMPTY = 0,      //!< 入力されていない
	E_SERCMD_ASCII_CMD_READCOLON,      //!< E_SERCMD_ASCII_CMD_READCOLON
	E_SERCMD_ASCII_CMD_READPAYLOAD,    //!< E_SERCMD_ASCII_CMD_READPAYLOAD
	E_SERCMD_ASCII_CMD_READCR,         //!< E_SERCMD_ASCII_CMD_READCR
	E_SERCMD_ASCII_CMD_READLF,         //!< E_SERCMD_ASCII_CMD_READLF
	E_SERCMD_ASCII_CMD_COMPLETE = 0x80,//!< 入力が完結した(LCRチェックを含め)
	E_SERCMD_ASCII_CMD_ERROR = 0x81,          //!< 入力エラー
	E_SERCMD_ASCII_CMD_CHECKSUM_ERROR = 0x82,       //!< LRCが間違っている
} teSercmdAsciiState;

sercmd::sercmd()
{
	bBinary = false;

	u16maxlen = MAX_LEN;
	au8data = new unsigned char[u16maxlen];
	vPutChar = NULL;
	u16timeout = 0;
}

sercmd::sercmd( bool bbin, unsigned short timeout, void (*cbPutChar)(unsigned char u8byte) )
{
	bBinary = bbin;

	u16maxlen = MAX_LEN;
	au8data = new unsigned char[u16maxlen];
	vPutChar = cbPutChar;
	u16timeout = timeout;
}

sercmd::~sercmd()
{
	delete au8data;
}

unsigned char sercmd::u8Parse( unsigned char u8byte )
{
	unsigned char u8Sta;
	if(bBinary){
		u8Sta = sercmd::u8BinaryParse(u8byte);
	}else{
		u8Sta = sercmd::u8AsciiParse(u8byte);
	}
	return u8Sta;
}

void sercmd::vOutput()
{
	if(bBinary){
		sercmd::vBinaryOutput();
	}else{
		sercmd::vAsciiOutput();
	}
}

/** @ingroup SERCMD
 * シリアルコマンド系列の解釈を行う状態遷移マシン
 *
 * @param pCmd
 * @param u8byte
 * @return
 */
unsigned char sercmd::u8BinaryParse(unsigned char u8byte) {
	// check for timeout
	if ( u16timeout && u8state != E_SERCMD_BINARY_EMPTY) {
		if ( millis() - u32timestamp > u16timeout ) {
			u8state = E_SERCMD_BINARY_EMPTY; // start from new
		}
	}

	// check for complete or error status
	if ( u8state >= 0x80 ) {
		u8state = E_SERCMD_BINARY_EMPTY;
	}

	// run state machine
	switch (u8state) {
	case E_SERCMD_BINARY_EMPTY:
		if (u8byte == SERCMD_SYNC_1) {
			u8state = E_SERCMD_BINARY_READSYNC;
			u32timestamp = millis(); // store received time
		}
		break;

	case E_SERCMD_BINARY_READSYNC:
		if (u8byte == SERCMD_SYNC_2) {
			u8state = E_SERCMD_BINARY_READLEN;
		} else {
			u8state = E_SERCMD_BINARY_ERROR;
		}
		break;

	case E_SERCMD_BINARY_READLEN:
		if (u8byte & 0x80) {
			// long length mode (1...
			u8byte &= 0x7F;
			u16len = u8byte;
			if (u16len <= u16maxlen) {
				u8state = E_SERCMD_BINARY_READLEN2;
			} else {
				u8state = E_SERCMD_BINARY_ERROR;
			}
		} else {
			// short length mode (1...127bytes)
			u16len = u8byte;
			if (u16len && u16len <= u16maxlen) {
				u8state = E_SERCMD_BINARY_READPAYLOAD;
				u16pos = 0;
				u16cksum = 0;
			} else {
				u8state = E_SERCMD_BINARY_ERROR;
			}
		}
		break;

	case E_SERCMD_BINARY_READLEN2:
		u16len = (u16len * 256) + u8byte;
		if (u16len && u16len <= u16maxlen) {
			u8state = E_SERCMD_BINARY_READPAYLOAD;
			u16pos = 0;
			u16cksum = 0;
		} else {
			u8state = E_SERCMD_BINARY_ERROR;
		}
		break;

	case E_SERCMD_BINARY_READPAYLOAD:
		au8data[u16pos] = u8byte;
		u16cksum ^= u8byte; // update XOR checksum
		if (u16pos == u16len - 1) {
			u8state = E_SERCMD_BINARY_READCRC;
		}
		u16pos++;
		break;

	case E_SERCMD_BINARY_READCRC:
		u16cksum &= 0xFF;
		if (u8byte == u16cksum) {
			u8state = E_SERCMD_BINARY_COMPLETE;
		} else {
			u8state = E_SERCMD_BINARY_CRCERROR;
		}
		break;

	default:
		break;
	}

	return u8state;
}

/** @ingroup SERCMD
 * バイナリ形式の出力 (vOutputメソッド用)
 * @param pc
 * @param ps
 */
void sercmd::vBinaryOutput() {
	int i;
	unsigned char u8xor = 0;

	if (vPutChar == NULL) return;

	vPutChar(SERCMD_SYNC_1);
	vPutChar(SERCMD_SYNC_2);
	vPutChar((unsigned char)(0x80 | (u16len >> 8)));
	vPutChar(u16len & 0xff);

	for (i = 0; i < u16len; i++) {
		u8xor ^= au8data[i];
		vPutChar(au8data[i]);
	}

	vPutChar(u8xor); // XOR check sum
	vPutChar(0x4); // EOT
}

/** @ingroup SERCMD
 * シリアルコマンドアスキー形式の入力系列の解釈を行う。\n
 * - :[0-9A-Z]+(CRLF) 系列の入力を行い、LRCチェックを行う
 *
 * @param pCmd 管理構造体
 * @param u8byte 入力文字
 * @return 状態コード (teModbusCmdState 参照)
 */
unsigned char sercmd::u8AsciiParse(unsigned char u8byte) {
	// check for timeout
	if (u16timeout && u8state != E_SERCMD_ASCII_CMD_EMPTY) {
		if ( millis() - u32timestamp > u16timeout ) {
			u8state = E_SERCMD_ASCII_CMD_EMPTY; // start from new
		}
	}

	// check for complete or error status
	if (u8state >= 0x80) {
		u8state = E_SERCMD_ASCII_CMD_EMPTY;
	}

	// run state machine
	switch (u8state) {
	case E_SERCMD_ASCII_CMD_EMPTY:
		if (u8byte == ':') {
			u8state = E_SERCMD_ASCII_CMD_READPAYLOAD;

			u32timestamp = millis(); // store received time for timeout
			u16pos = 0;
			u16cksum = 0;
			u16len = 0;
			// memset(pCmd->au8data, 0, pCmd->u16maxlen);
		}
		break;

	case E_SERCMD_ASCII_CMD_READPAYLOAD:
		if ((u8byte >= '0' && u8byte <= '9') || (u8byte >= 'A' && u8byte <= 'F')) {
			/* オーバーフローのチェック */
			if (u16pos / 2 == u16maxlen) {
				u8state = E_SERCMD_ASCII_CMD_ERROR;
				break;
			}

			/* 文字の16進変換 */
			const unsigned char u8tbl[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0, 0, 10, 11, 12, 13, 14, 15 };
			unsigned char u8val = u8tbl[u8byte - '0'];

			/* バイナリ値として格納する */
			unsigned char *p = &au8data[u16pos / 2];

			if (u16pos & 1) {
				*p = (*p & 0xF0) | u8val;
				u16cksum += *p;
			} else {
				*p = u8val << 4;
			}

			u16pos++;
		} else if (u8byte == 0x0d || u8byte == 0x0a) { // CR入力
			if (u16pos >= 4 && ((u16pos & 1) == 0) // データ部１バイト、チェックサム１バイト以上
			) {
				// チェックサムの確認
				u16cksum &= 0xFF; // チェックサムを 8bit に切り捨てる
				if (u16cksum) { // 正しければ 0 になっているはず
					// 格納値
					unsigned char u8lrc = au8data[u16pos / 2 - 1]; // u16posは最後のデータの次の位置
					// 計算値(二の補数の計算、ビット反転+1), デバッグ用に入力系列に対応する正しいLRCを格納しておく
					u16cksum = (~(u16cksum - u8lrc) + 1) & 0xFF;
					u8state = E_SERCMD_ASCII_CMD_CHECKSUM_ERROR;
				} else {
					// LRCが正しければ、全部足したら 0 になる。
					u8state = E_SERCMD_ASCII_CMD_COMPLETE; // 完了！
					u16len = u16pos / 2 - 1;
				}
			} else {
				u8state = E_SERCMD_ASCII_CMD_ERROR;
			}
		} else if (u8byte == 'X') {
			// X で終端したらチェックサムの計算を省く
			if (u16pos >= 2 && ((u16pos & 1) == 0)) { // 入力データあり
				u8state = E_SERCMD_ASCII_CMD_COMPLETE; // 完了！
				u16len = u16pos / 2;
			}
		} else {
			u8state = E_SERCMD_ASCII_CMD_EMPTY;
		}
		break;

	default:
		break;
	}

	return u8state;
}

/** @ingroup SERCMD
 * シリアルコマンドアスキー形式の出力補助関数。１バイトを１６進数２文字で出力する (0xA5 -> "A5")
 *
 * @param psSerStream 出力先
 * @param u8byte 出力したいバイト列
 * @param pu8lrc LRC計算用
 */
inline void sercmd::vPutByte(unsigned char u8byte, unsigned char *pu8lrc) {
	static const unsigned char au8tbl[] = "0123456789ABCDEF";
	unsigned char u8OutH, u8OutL;

	if (vPutChar == NULL) return;

	u8OutH = au8tbl[u8byte >> 4];
	u8OutL = au8tbl[u8byte & 0x0F];

	vPutChar(u8OutH);
	vPutChar(u8OutL);

	if (pu8lrc) {
		*pu8lrc += u8byte;
	}
}

/** @ingroup SERCMD
 * アスキー形式の出力 (vOutputメソッド用)
 * @param pc
 * @param ps
 */
void sercmd::vAsciiOutput() {
	int i;
	unsigned char u8lrc = 0;

	if (vPutChar == NULL) return;

	// 先頭の :
	vPutChar(':');

	for (i = 0; i < u16len; i++) {
		vPutByte(au8data[i], &u8lrc);
	}

	// ２の補数の計算 (LRC)
	u8lrc = ~u8lrc + 1;
	vPutByte(u8lrc, NULL);

	// Trailer byte
	vPutChar(0x0D);
	vPutChar(0x0A);
}
